/* Copyright (c) 2001 AT&T Corp.  All rights reserved.

This software may only be used by you under license from AT&T Corp. 
("AT&T").  A copy of AT&T's Source Code Agreement is available at 
AT&T's Internet website having the URL 

http://www.research.att.com/sw/tools/graphviz/license/

If you received this software without first entering into a license 
with AT&T, you have an infringing copy of this software and cannot 
use it without violating AT&T's intellectual property rights. */

#ifndef Dynagraph_h
#define Dynagraph_h

#include "LGraph-cdt.h"
#include "traversal.h"
#include "StrAttr.h"
#include "Geometry.h"


/*
	UPDATE flags. use with ChangeQueue::ModNode,ModEdge to set this 
	modify subgraph specific flag.  
	(this update flag is a poor use for instance-specific data, because the data 
	is unused in the other subgraphs)
*/
typedef enum {
	DG_UPD_MOVE = 1<<0, // NodeGeom:: or EdgeGeom:: pos attribute
	DG_UPD_REGION = 1<<1, // NodeGeom::region
	DG_UPD_NAIL = 1<<2, // NodeGeom::nail
	DG_UPD_TAIL = 1<<3, // EdgeGeom::tailPort or ::tailClipped 
	DG_UPD_HEAD = 1<<4, // (sorry - can't change end nodes! create a new edge and delete this one)
	DG_UPD_WIDTH = 1<<5, // EdgeGeom::
	DG_UPD_LENGTH = 1<<6, // EdgeGeom::
	DG_UPD_COST = 1<<7, // EdgeGeom::
	DG_UPD_CONSTRAINT = 1<<8, // EdgeGeom::
	DG_UPD_LABEL = 1<<9, // all Geom:: any label changed (hmm this update flag model is breaking down.  should i reserve 5 bits for 32 labels har har)
	DG_UPD_BOUNDS = 1<<10, // NodeGeom:: and GraphGeom::
	DG_UPD_NESTING = 1<<11, // dream on
	DG_UPD_EDGESTYLE = 1<<12, // GraphGeom::splineLevel (useful for debugging.  real edge styles also should be on the way...)
	DG_UPD_RESOLUTION = 1<<13, // GraphGeom::
	DG_UPD_SEPARATION = 1<<14 // GraphGeom::
} UpdateFlags;
struct Update { // subgraph-specific datum
	unsigned flags;
	Update(unsigned flags = 0) : flags(flags) {}
};

/*
	MODEL pointer - annoying type unsafe corner in each layout object for particular layout servers to put back pointers

	Really, the layout servers should be templated on the layout graph type:
	a server can insist on whatever data it needs (typesafe) just by asking for it with gd<>

	But I haven't seen a great need yet, and compilation time for debugging the layout servers has been a factor.
*/
struct ModelPointer {
	void *model;
	ModelPointer() : model(0) {}
};

/*
	GRAPH attributes
*/
typedef enum {
	DG_SPLINELEVEL_VNODE,
	DG_SPLINELEVEL_BOUNDS,
	DG_SPLINELEVEL_SHORTEST,
	DG_SPLINELEVEL_SPLINE
} SpliningLevel;
struct GraphGeom {
	Bounds bounds;
	SpliningLevel splineLevel;
	Coord labelGap; // between node and label
	Coord resolution; // smallest units in each dimension (e.g. 1,1 for integer)
	Coord separation; // space between horizontally adjacent nodes and edges; minimum edge height
	float timeLimit; // in seconds (NYI in dynadag); 0 - no limit
	GraphGeom() : splineLevel(DG_SPLINELEVEL_SPLINE),labelGap(0,0),resolution(0.1,0.1),separation(0.5,0.5),timeLimit(0) {}
};
struct StaticLabel {
	Rect bounds;
};
struct StaticLabels {
	vector<StaticLabel> labels;
};
// this stuff is a stab at the higher-level management of queues and events that has to 
// happen with nesting.  not pretty.
struct Server;
struct ChangeQueue;
struct Management {
	Server *engine;
	ChangeQueue *queue;
	int locks;

	Management() : engine(0),queue(0),locks(0) {}
	void Lock(bool lock);
	void Go();
};
struct Client {
	Server *client;
	Client() : client(0) {}
};
struct GraphAttrs : Name,StrAttrs,StrChanges,GraphGeom,StaticLabels,ModelPointer,Client,Hit {};

/*
	NODE attributes
*/
typedef enum {
	DG_NONAIL = 0,
	DG_NAIL_X = 1,
	DG_NAIL_Y = 2,
	DG_NAIL_BOTH = 3
} NailType;
struct NodeGeom {
	Position pos;
	Region *region;
	NailType nail;
	NodeGeom() : region(0),nail(DG_NONAIL) {}
	~NodeGeom() {
		if(region) {
			delete region;
			region = 0;
		}
	}
	Bounds BoundingBox() {
		if(!pos.valid)
			return Bounds();
		if(region)
			return Bounds(Rect(region->boundary+pos));
		else 
			return Bounds(Rect(pos));
	}
};
typedef enum {
	DG_NODELABEL_CENTER,
	DG_NODELABEL_LEFT,
	DG_NODELABEL_TOP,
	DG_NODELABEL_RIGHT,
	DG_NODELABEL_BOTTOM
} NodeLabelPlacement;

struct NodeLabel {
	// input
	NodeLabelPlacement where;
	Coord size;
	// output
	Position pos; // upper-left corner

	NodeLabel(NodeLabelPlacement where = DG_NODELABEL_CENTER, Coord size = Coord(0,0)) :
		where(where), size(size) {}
};
struct Port {
	Coord pos;
	double angle;
	Port(Coord pos=Coord(),double angle=0) : pos(pos),angle(angle) {}
};
typedef vector<pair<int,Port> > IdentifiedPorts;
// implemented by both container and nesting
// additions and deletions get dealt with in implementation-specific ways
// when notifying a nesting, the ports are points along the boundary of the nesting
// when notifying upward, the ports are the end locations of the edges
struct Nesting {
	Nesting *notify; // the other end
	Nesting() : notify(0) {}
	virtual void Change(const IdentifiedPorts &what) = 0;
	virtual void Resize(Coord size) = 0;
};
struct NestedGraph {
	Nesting *n;
	NestedGraph() : n(0) {}
};
typedef vector<NodeLabel> NodeLabels;
struct NodeAttrs : Name,StrAttrs,StrChanges,NodeGeom,NodeLabels,NestedGraph,ModelPointer,Hit {};

/*
	EDGE attributes
*/
struct EdgeGeom {
	double width,lengthHint,cost;
	Line pos;
	bool constraint, // whether should affect ranking
		manualRoute; // try to use 
	Port tailPort,headPort;
	bool tailClipped,headClipped;
	EdgeGeom() : width(1.0),lengthHint(1.0),cost(0.0),constraint(true),
		manualRoute(false),tailPort(Coord(0.0,0.0)),headPort(Coord(0.0,0.0)),
		tailClipped(true),headClipped(true) {}
};
struct EdgeLabel {
	// input
	float where; // 0 at tail, 1 at head
	Rect shape; // offsets of label bounds from point on edge
	// output
	Position pos;
	EdgeLabel(float where = 0.5f, Rect shape = Rect(0,0,0,0)) : where(where), shape(shape) {}
};
struct InwardEdge {
	int tailID,headID; // used by NestingManager to communicate with nesting
	InwardEdge() : tailID(0),headID(0) {}
};
typedef vector<EdgeLabel> EdgeLabels;
struct EdgeAttrs : Name,StrAttrs,StrChanges,EdgeGeom,EdgeLabels,InwardEdge,ModelPointer,Hit {};

/*
	LAYOUT (spec & geom graph)
*/
typedef LGraph<GraphAttrs,NodeAttrs,EdgeAttrs,Update,Update,Update> Layout;

/*
	CHANGE QUEUE
*/
struct ChangeQueue {
	// the client edits this supergraph of server's current layout
	// then calls the methods below to signal the changes in the subgraphs
	Layout * const client, * const current; 
	Layout insN,modN,delN,
		insE,modE,delE;

	ChangeQueue(Layout *client,Layout *current);
	void InsNode(Layout::Node *n);
	void InsEdge(Layout::Edge *e);
	void ModNode(Layout::Node *n,Update u); // ORs u into the igd<Update> of modN subnode
	void ModEdge(Layout::Edge *e,Update u);
	void DelNode(Layout::Node *n);
	void DelEdge(Layout::Edge *e);
	unsigned &GraphUpdateFlags() { return igd<Update>(&modN).flags; }

	// called by server to update current subgraph based on current changes
	void UpdateCurrent();

	// called by client after server processing clear subgraphs and maybe do deletions
	void Okay(bool doDelete = false);

	bool Empty() { return insN.nodes().empty()&&modN.nodes().empty()&&delN.nodes().empty()&&
		insE.nodes().empty()&&modE.nodes().empty()&&delE.nodes().empty(); }

	// copy
	ChangeQueue &operator=(ChangeQueue &Q);
	// accumulate
	ChangeQueue &operator+=(ChangeQueue &Q);

	// Exceptions

	// insertions must not already be inserted; modifications & deletions must already be inserted
	struct InsertInserted {};
	struct ModifyUninserted {};
	struct DeleteUninserted {};
	// insertion of edge without nodes
	struct EndnodesNotInserted {};
};
/*
	SERVER (e.g. topology layout, edge layout, label layout)
*/
struct Server {
	// shared by all actors in a system;  represents everything currently inserted
	// note this is a subgraph of the client layout
	Layout * const current; 

	virtual void Process(ChangeQueue &Q) = 0;
	Server(Layout *current) : current(current) {}
	virtual ~Server() {}
};
struct CompoundServer : Server {
	typedef vector<Server*> ServerV;
	ServerV actors;
	void Process(ChangeQueue &Q);
	CompoundServer(Layout *clientLayout) : Server(clientLayout) {}
	~CompoundServer();
};
struct LabelPlacer : Server {
	void Process(ChangeQueue &Q);
	LabelPlacer(Layout *clientLayout) : Server(clientLayout) {}
	~LabelPlacer() {}
};
/*
// servers should look like this:
struct LayoutServer : Server {
	LayoutServer(Layout *clientLayout) : 
		Server(clientLayout) {}
};
*/

/* 
	EXCEPTIONS
*/
struct NailWithoutPos {
	Layout::Node *n;
	NailWithoutPos(Layout::Node *n) : n(n) {}
};

#endif
