/* Copyright (c) 2001 AT&T Corp.  All rights reserved.

This software may only be used by you under license from AT&T Corp. 
("AT&T").  double copy of AT&T's Source Code Agreement is available at 
AT&T's Internet website having the URL 

http://www.research.att.com/sw/tools/graphviz/license/

If you received this software without first entering into a license 
with AT&T, you have an infringing copy of this software and cannot 
use it without violating AT&T's intellectual property rights. */

#ifndef Geometry_h
#define Geometry_h

#pragma warning (disable : 4786 4503)
#include <math.h>
#include "moremath.h"
#include <vector>
#include <list>
#include <float.h>
#include "useful.h"
using namespace std;
// sorry.  I do not understand this
#ifdef _MSC_VER
#include <minmax.h>
#endif

struct Coord {
	double x,y;
	Coord() {}
	Coord(double x, double y) : x(x),y(y){}
	bool operator ==(const Coord &c) const {
		return double(x)==double(c.x) && double(y)==double(c.y);
	}
	bool operator !=(const Coord &c) const {
		return !(*this==c);
	} 
	Coord operator +(const Coord &c) const {
		return Coord(x+c.x,y+c.y);
	}
	Coord operator -(const Coord &c) const {
		return Coord(x-c.x,y-c.y);
	}
	Coord operator -() const {
		return Coord(-x,-y);
	}
	Coord operator *(double a) const {
		return Coord(a*x,a*y);
	}
	Coord operator /(double a) const {
		return Coord(x/a,y/a);
	}
	Coord &operator +=(const Coord &c) {
		return *this = *this + c;
	}
	Coord &operator -=(const Coord &c) {
		return *this = *this - c;
	}
	Coord &operator *=(double a) {
		return *this = *this*a;
	}
	Coord &operator /=(double a) {
		return *this = *this/a;
	}
	// dot product
	double operator *(Coord a) const {
		return x*a.x+y*a.y;
	}
	double Len() const {
		return sqrt(*this * *this);
	}
	Coord Norm() const {
		double d = Len();
		if(d)
			return *this / d;
		else
			return *this;
	}
	Coord Abs() const {
		return Coord(fabs(x),fabs(y));
	}
};
inline double dSquared(Coord a,Coord b) {
	Coord c = a-b;
	return c*c;
}
inline double dist(Coord a,Coord b) {
	return sqrt(dSquared(a,b));
}
struct Position : Coord {
	bool valid;
	Position() : valid(false), Coord(-17,-17) {}
	Position(double x, double y) : valid(true), Coord(x,y) {}
	Position(Coord c) : valid(true), Coord(c) {}
	Position &operator =(const Coord &c) {
		static_cast<Coord&>(*this) = c;
		valid = true;
		return *this;
	}
	void invalidate() {
		valid = false;
		x = y = -17;
	}
	bool operator ==(const Position &p) const {
		return valid==p.valid && ((Coord)p==*this);
	}
	bool operator !=(const Position &p) const {
		return !(*this==p);
	}
};
struct Segment {
	Coord a,b;
	Segment(Coord a = Coord(),Coord b = Coord()) : a(a), b(b) {}
};
struct Rect {
	double l,t,r,b;
	Rect() {}
	Rect(const Rect &r) : l(r.l),t(r.t),r(r.r),b(r.b) {}
	Rect(double l,double t,double r,double b) : l(l),t(t),r(r),b(b) {}
	Rect(Coord c) : l(c.x),t(c.y),r(c.x),b(c.y) {}
	bool operator==(const Rect &r) const {
		return l==r.l && t==r.t && this->r==r.r && b==r.b;
	}
	bool operator!=(const Rect &r) const {
		return !(*this==r);
	}
	double Width() const {
		return r-l;
	}
	double Height() const {
		return fabs(t-b);
	}
	Coord Size() const {
		return Coord(Width(),Height());
	}
	bool Contains(Coord c) const {
		return l<=c.x && c.x<=r &&
			(b<t)?(b<=c.y && c.y<=t):(t<=c.y && c.y<=b);
	}
	Coord Center() const {
		return Coord((l+r)/2.0,(t+b)/2.0);
	}
	Coord LowerLeft() const {
		return Coord(l,b);
	}
	Coord UpperRight() const {
		return Coord(r,t);
	}
	Rect operator+(Coord c) const {
		return Rect(l+c.x,t+c.y,r+c.x,b+c.y);
	}
	Rect operator-(Coord c) const {
		return Rect(l-c.x,t-c.y,r-c.x,b-c.y);
	}
};
struct Bounds : Rect {
	bool valid;
	Bounds() : valid(false) {}
	Bounds(Rect r) : Rect(r),valid(true) {}
	bool operator ==(const Bounds &b) const {
		return !valid&&!b.valid || Rect::operator ==(b);
	}
	bool operator !=(const Bounds &b) const {
		return !(*this==b);
	}
};
typedef enum {DG_RECT_REGION,DG_POLYLINE_REGION,DG_ELLIPSE_REGION} RegionType; // to make downcasts safe
struct Region {
	RegionType type;
	Bounds boundary; // perhaps bounding poly instead, for better spline routing?
	Region(RegionType type=DG_RECT_REGION) : type(type) {}
	virtual ~Region() {}
	virtual bool Hit(Coord c) const {
		return boundary.Contains(c);
	}
};
struct Line : vector<Coord > {
	typedef vector<Coord >::iterator iterator; // for gcc 3.0; why?  
    Coord &at(int i) { return operator[](i); } // for earlier gccs
	int degree;
	Line() : degree(0) {}
	void Clear() {
		clear();
		degree = 0;
	}
	bool Empty() {
		return degree==0 || !size();
	}
	int GetSeg(double y);
	template<typename It> // T* and vector<T>::iterator are finally different types in gcc 3.0!
	void AddSeg(It ci) {
		It b = ci,
			e = ci+degree+1;
		if(size()) {
			assert(back()==*b);
			b++;
		}
		insert(end(),b,e);
	}
	Position YIntersection(double y);
	Position Intersection(int seg,Segment other);
	Line &operator +=(const Line &append) {
		assert(degree==append.degree);
		Coord b = back(),f = append.front();
		assert(b==f);
		insert(end(),append.begin()+1,append.end());
		return *this;
	}
	void ClipEndpoints(Line &source,const Coord offsetTail,const Region *tl,
		const Coord offsetHead,const Region *hd); // NG*'s NULL to not clip
private:
	Coord *clip(Coord *in,const Coord offset,const Region &reg,Coord *out);
};
typedef list<Line> Lines;

// check that YIntersection exists
inline Coord checkPos(Position p) { check(p.valid); return p; }

struct PolylineRegion : Region {
	Line shape;
	mutable int lastOut; // to make iterative calls quicker
	PolylineRegion() : lastOut(0),Region(DG_POLYLINE_REGION) {}
	~PolylineRegion() {}
	void update(); // update Bounds from new Line info
	bool Hit(Coord c) const; // Region impl
private:
	bool sameSide(const Coord &p0, const Coord &p1, int seg) const;
};
struct EllipseRegion : Region {
	EllipseRegion() : Region(DG_ELLIPSE_REGION) {}
	~EllipseRegion() {}
	bool Hit(Coord c) const {
		// (x/a)^2 + (y/b)^2 < 1
		Coord O = boundary.Center(),
			c2 = c-O;
		c2.x /= boundary.Width()/2;
		c2.y /= boundary.Height()/2;
		return c2*c2 < 1;
	}
};
#include "GeomUtil.h"
#include "LineTricks.h"
#include "StreamGeom.h"

#endif // Geometry_h
