/* Copyright (c) 2001 AT&T Corp.  All rights reserved.

This software may only be used by you under license from AT&T Corp. 
("AT&T").  A copy of AT&T's Source Code Agreement is available at 
AT&T's Internet website having the URL 

http://www.research.att.com/sw/tools/graphviz/license/

If you received this software without first entering into a license 
with AT&T, you have an infringing copy of this software and cannot 
use it without violating AT&T's intellectual property rights. */

#include "Dynagraph.h"

// implementation of nesting for Dynagraph where the engines can remain ignorant of nesting

// wraps a subServer manager and deals with things nested within
struct NestingContainer : Server {
	void Process(ChangeQueue &Q);
	NestingContainer(Server *l) : subServer(l),Server(l->current),watching(l->current),callbackQ(l->current->parent,l->current),callbackDepth(0) {}
	~NestingContainer();
private:
	Server *subServer;
	Layout watching;
	// there are three circumstances in which calls can come in from children:
	// 0. a change that originated in the child: respond immediately
	// 1. callback from child's Nesting::Change: queue the change
	// 2. callback that's a result of a callback: ignore the change
	int callbackDepth;
	struct Callback : Nesting {
		Layout::Node *n;
		typedef map<int,Layout::Edge*> IdentifiedEdges;
		IdentifiedEdges edges;
		NestingContainer *container;
		Callback() : n(0),container(0) {}
		void Change(const IdentifiedPorts &what);
		void Resize(Coord size);
	};
	friend struct Callback;
	typedef map<Layout::Node*,Callback*> Callbacks;
	Callbacks callbacks;
	ChangeQueue callbackQ;
	void doCallback(Layout::Node *n,Nesting *other);
	void updateEdgeCallbacks(Layout::Edge *e);
	void findNestingUpdates(ChangeQueue &Q);
	void sendChangesIn(ChangeQueue &Q);
	void round2(); // the second outer and inner subServer
};
// wraps a subServer and implements Nesting for the container
// note that this uses its own ChangeQueue when calling into the subServer.
// so if anyone needs to know of the changes made from here, they should implement
// Server and wrap the subServer's implementation.
struct NestedDynagraph : Nesting,Server {
	NestedDynagraph(Server *l) : Server(l->current),subServer(l),currPort(0),callbackDepth(0),callbackQ(l->current->parent,l->current) {}
	~NestedDynagraph() { delete subServer; }
// insert and delete.
	pair<int,Layout::Edge*> AddOutward(Layout::Node *n,bool isHead);
	void DeleteOutward(int ID);

// bad, bad, bad
	ChangeQueue callbackQ;
	void FlushQ() {
		Process(callbackQ); 
		if(callbackDepth==0)
			gd<Client>(current).client->Process(callbackQ);
		callbackQ.Okay();
	}
// Nesting:
	void Change(const IdentifiedPorts &what);
	void Resize(Coord size);
// Server
	void Process(ChangeQueue &Q);
private:
	// calls to Change can result from 
	typedef map<int,Layout::Node*> BoundaryNodes;
	BoundaryNodes boundaryNodes;
	int currPort;
	Server *subServer;
	int callbackDepth;
	void doCallback(ChangeQueue &Q);
};

inline NestedDynagraph *nest(Server *engine) {
	return new NestedDynagraph(new NestingContainer(engine));
}
struct CombinedEdge {
	CombinedEdge(Layout::Node *tail,NestedDynagraph *man) : curr(tail),man(man),ID(0) {}
	// build...
	// the only valid call sequence is zero-or-more Leave, then a Connect, then zero-or-more Enter
	Layout::Edge *Leave(NestedDynagraph *outerLayout,Layout::Node *outerNode) {
		pair<int,Layout::Edge*> res = man->AddOutward(curr,false);
		segments.push_back(Segment(res.second,res.first,man));
		if(ID) 
			gd<InwardEdge>(res.second).tailID = ID;
		ID = res.first;
		man = outerLayout;
		curr = outerNode;
		return res.second;
	}
	Layout::Edge *Connect(Layout::Node *head) {
		Layout::Edge *e = man->current->parent->create_edge(curr,head).first;
		man->callbackQ.InsEdge(e);
		segments.push_back(Segment(e,0,man));
		if(ID) {
			gd<InwardEdge>(e).tailID = ID;
			ID = 0;
		}
		curr = head;
		man = 0;
		return e;
	}
	Layout::Edge *Enter(NestedDynagraph *innerLayout,Layout::Node *innerNode) {
		assert(!ID);
		assert(!man);
		assert(!segments.empty());
		pair<int,Layout::Edge*> res = innerLayout->AddOutward(innerNode,true);
		gd<InwardEdge>(segments.back().edge).headID = res.first;
		segments.push_back(Segment(res.second,res.first,innerLayout));
		curr = innerNode;
		return res.second;
	}
	// go
	void Insert() { 
		for(Segments::iterator si = segments.begin(); si!=segments.end(); ++si)
			if(!si->ID)
				break;
		assert(si!=segments.end());
		si->nested->FlushQ();
	}
	void GetLine(Line &out) {
		out.Clear();
		out.degree = gd<EdgeGeom>(segments.front().edge).pos.degree;
		for(Segments::iterator si = segments.begin(); si!=segments.end(); ++si)
			out += gd<EdgeGeom>(si->edge).pos;
	}
	void Delete() {
		for(Segments::iterator si = segments.begin(); si!=segments.end(); ++si) {
			if(si->ID) 
				si->nested->DeleteOutward(si->ID);
			else
				si->nested->callbackQ.DelEdge(si->edge);
			si->nested->FlushQ();
		}
	}
private:
	struct Segment {
		Layout::Edge *edge;
		int ID;
		NestedDynagraph *nested;
		Segment(Layout::Edge *edge,int ID,NestedDynagraph *nested) : edge(edge),ID(ID),nested(nested) {}
	};
	typedef vector<Segment> Segments;
	Segments segments;
	Layout::Node *curr;
	NestedDynagraph *man;
	int ID;
};